import Component from '@glimmer/component';
import { tracked } from '@glimmer/tracking';
import moment from 'moment';
import { bisector } from "d3-array";
import { throttle } from '@ember/runloop';

export default class MultiScrubberComponent extends Component {
  get bisectDate() {
    return bisector(d => d.at).left;
  }

  @tracked fadeState = "opacity-0";
  @tracked currentMousePos;
  @tracked cornerTextValues;
  @tracked line;
  @tracked movingTextContainer;
  @tracked textBox;
  @tracked animatedTextXPosition;
  @tracked textSizeElement;
  @tracked hoverTextValues;

  get strokeWidth() {
    return this.args.width ? this.args.width : '1';
  }

  get lineColor() {
    return this.args.lineColor || 'gray';
  }

  get lineHeight() {
    return this.args.chart.getBoundingClientRect().height;
  }

  get lineTextUnits() {
    return this.args.lineTextUnits || '';
  }

  get cornerTextUnits() {
    return this.args.cornerTextUnits || '';
  }

  get lineTextKey() {
    return this.args.lineTextKey || 'value';
  }

  get cornerTextKey() {
    return this.args.cornerTextKey || 'at';
  }

  get singleValueCornerText() {
    return this.args.singleValueCornerText || false;
  }

  get primaryDataset() {
    return this.args.datasets ? this.args.datasets[0] : [];
  }

  get boxYLocation() {
    return this.args.boxYLocation || 40;
  }

  get textBoxFillOpacity() {
    return this.args.textBoxFillOpacity || '1';
  }

  get textBoxFill() {
    return this.args.textBoxFill || 'white';
  }

  get textLineHeight() {
    return this.args.textLineHeight || 15
  }
  
  get isMultiScale() {
    return typeof this.args.xScale === 'object';
  }

  get textBoxPadding() {
    return this.args.textBoxPadding || 20;
  }

  calculateText(value) {
    if( value == null ) return "";
    if(value instanceof Date) {
      return this.formatDate(value);
    } else {
      return `${parseFloat(value.toFixed(2))}${this.lineTextUnits}`; 
    }
  }

  get hasLineText() {
    return this.hoverTextValues && this.hoverTextValues.length > 0;
  }

  get cornerTexts() {
    if (!this.cornerTextValues) return []
    return this.cornerTextValues.map((value) => {
      if(value instanceof Date) {
        return this.formatDate(value);
      } else {
        return `${value}${this.cornerTextUnits}`;
      }
    });
  }

  updateLine = (e) => {
    const mousePos = this.mouseOrTouchPosition(e);
    this.currentMousePos = mousePos;
    if( mousePos > 0.5 ) {
      const nearestPoints = this.findNearestDatapoints(mousePos);
      const yLoc = this.findMin(nearestPoints, 'y');
      this.textBox = this.findTextXLocBoundedBySvg(nearestPoints, yLoc);
      this.hoverTextValues = nearestPoints.map((datapoint, index) => {
        return {
          ...datapoint,
          y: this.textBox.y + this.textBoxPadding + (index * this.textLineHeight + 10),
          x: this.textBox.x + this.textBoxPadding
        };
      });

      this.cornerTextValues =  this.singleValueCornerText
          ? [ nearestPoints[0][this.cornerTextKey] ]
          : nearestPoints.map((d) => d[this.cornerTextKey]);
    }
  }

  findMin(array, field) {
    return array.reduce((max, obj) => {
      return obj && obj[field] < max ? obj[field] : max;
    }, this.args.yScale(0));
  }

  formatDate(date) {
    const dateObj = new Date(date);
    const hours = dateObj.getHours();
    const minutes = dateObj.getMinutes();
    const time = `, ${hours < 12 ? hours : hours - 12}:${('0' + minutes).slice(-2)} ${hours < 12 ? "am" : "pm"}`
    return `${dateObj.getMonth()+1}-${dateObj.getDate()}-${dateObj.getFullYear()}${this.hideTime ? "" : time}`;
  }

  findNearestDatapoints(mousePos) {
    if (!this.args.datasets) return [{x: "", y: ""}];

    return this.args.datasets.map((dataset, index) => {
      const xScale = this.isMultiScale ? this.args.xScale[index] : this.args.xScale;
      const x0 = xScale.invert(mousePos); // x position within the element.
      const dataPoint = this.nearestDataPoint(dataset, x0);
      const x = dataPoint ? this.args.xScale(dataPoint[this.cornerTextKey]) : 0;
      return {
        ...dataPoint,
        x: x < 0 ? 0 : x,
        y: dataPoint ? this.args.yScale(dataPoint[this.lineTextKey]) - this.textLineHeight : 0,
        color: this.args.colors ? this.args.colors[index] : 'black',
        text: this.calculateText(dataPoint ? dataPoint[this.lineTextKey] : null)
      }
    });
  }

  nearestDataPoint(dataset, x0) {
    const i = this.bisectDate(dataset, x0, 1);
    const d0 = dataset[i - 1];
    const d1 = dataset[i];
    // out of dataset bounds
    if(dataset.length === 0 || x0 < dataset[0].at || x0 > dataset[dataset.length - 1].at) return
    return (d1 && d0) ? (x0 - d0.at > d1.at - x0 ? d1 : d0) : 0;
  }

  findTextXLocBoundedBySvg(dataPoints, yLoc) {
    if(!dataPoints.length) return 0
    const pointWithLargestText = dataPoints.reduce((max, dataPoint) => !max || dataPoint.text.length > max ? dataPoint : max);
    const textSize = this.calculateTextDimensions(pointWithLargestText.text);
    const textBoxWidth = this.textBoxPadding * 2 + textSize.width
    const barCenterPoint = dataPoints[0].x + (this.barWidth/2);
    const leftSide = barCenterPoint - (textBoxWidth)/2 ;
    const rightSide = dataPoints[0].x + textSize.width + this.textBoxPadding;
    const y = yLoc - (this.textBoxPadding * 4 + (this.textLineHeight * dataPoints.length))
    return {
      x: leftSide < 5 ? 5 : rightSide > (this.chartWidth - 5) ? this.chartWidth - textBoxWidth - 5 : leftSide,
      y: y < 5 ? 5 : y,
      width: textBoxWidth,
      height: this.textLineHeight * dataPoints.length + this.textBoxPadding * 2,
      translate: textBoxWidth / 2
    }
  }  

  mouseOrTouchPosition(event) {
    const clientX = event.type.includes("touch") ? event.targetTouches[0].clientX : event.clientX;
    return clientX - this.args.chart.getBoundingClientRect().left;
  }

  mousemove = (e) => {
    throttle(this, this.updateLine, e, 20);
  }

  mouseout = () => {
    this.fadeState = "opacity-0";
  }

  mouseover = (e) => {
    if(e.type === "touchstart") {
      this.currentMousePos = this.mouseOrTouchPosition(e);
      this.setMovingElementsPosition(this.currentMousePos);
    }
    this.fadeState = "opacity-100";
  }

  initElement = (element, [elementName]) => {
    this[elementName] = element;
  }

  animateScrubbing = () => {
    const x = (this.currentMousePos || this.currentMousePos === 0)
            ? this.getAnimatedLocation(this.line, 'x1')
            : 0
    if(x) {
      this.setMovingElementsPosition(x);
    }
  }

  setMovingElementsPosition = (x) => {
    this.line.setAttribute('x1', x)
    this.line.setAttribute('x2', x)
    this.movingTextContainer.setAttribute('x', x)
    this.animatedTextXPosition = x;
  }

  getAnimatedLocation(element, attrName) {
    const currentX = parseInt(element.getAttribute(attrName)); // Get current position
    const totalDx = this.currentMousePos - currentX; // Find total x distance from cursor to element
    const acceleration = 4; // acceleration scalar
    const frameDx = totalDx / acceleration; // delta x for this frame

    return Math.abs(frameDx) > 1                      // if the delta x is less than 1 it needs to be
            ? currentX + frameDx                      // rounded or else it will register as 0px and get stuck
            : totalDx < 1 ? this.currentMousePos : currentX + 1; // note only happens really close to cursor as it deccelerates
  }

  calculateTextDimensions(text) {
    this.textSizeElement.innerHTML = text;
    return {
      width: this.textSizeElement.clientWidth + 1,
      height: this.textSizeElement.clientHeight + 1
    }
  }

}